/* @(#) $Id: ifsgr.h,v 1.37 2004/05/13 14:04:28 yeti Exp $ */
#ifndef __IFSGR_H__
#define __IFSGR_H__ 1

#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

#define IFS_EPSILON 1e-9

#define IFS_PROBSTACK_BITS 11
#define IFS_PROBSTACK (1UL << (IFS_PROBSTACK_BITS))
#define IFS_PROBSTACK_MASK (IFS_PROBSTACK - 1UL)

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

typedef struct {
    double x, y;
} PointXY;

typedef struct {
    double x, y;
    double xx, xy, yx, yy;
    double p;
} IFSTrans;

typedef struct {
    char     *name;
    size_t    n;
    IFSTrans *t;
} IFS;

typedef struct {
    double  x, y;
    double  theta;
    double  scale;
    double  asym;
    double  shear;
    int     flip;
    double  prob;
} IFSCTrans;

IFS*       ifs_read_next              (FILE *fh);
IFS**      ifs_read_file              (FILE *fh,
                                       size_t *nifs);
IFS*       ifs_read_by_name           (FILE *fh,
                                       const char *name);
void       ifs_write_ifs              (const IFS *ifs,
                                       FILE *fh);
void       ifs_export_ifscompose      (const IFS *ifs,
                                       FILE *fh);
IFSCTrans* ifs_read_ifscompose        (FILE *fh,
                                       PointXY *center,
                                       size_t *ntrans);
IFS*       ifs_read_fdesign           (FILE *fh);
void       ifs_free                   (IFS *ifs);
void       ifs_normalize_probabilities(IFS *ifs);
PointXY    ifs_find_center            (const IFS *ifs);
PointXY    ifs_trans_get_fixed_point  (const IFSTrans *t);
PointXY    ifs_trans_get_scales       (const IFSTrans *t);
PointXY    ifs_trans_get_rotations    (const IFSTrans *t);
bool       ifs_trans_is_contraction   (const IFSTrans *t);
IFSTrans*  ifs_trans_rsr_compose      (PointXY pq,
                                       PointXY phi,
                                       IFSTrans *t);
void       ifs_find_range             (const IFS *ifs,
                                       const size_t *tab,
                                       uint64_t niter,
                                       PointXY *min,
                                       PointXY *max);
/* broken
void       ifs_find_range_extrapolated(const IFS *ifs,
                                       const size_t *tab,
                                       uint64_t niter,
                                       PointXY *min,
                                       PointXY *max);
                                       */
void       ifs_fix_transforms         (IFS *ifs,
                                       PointXY min,
                                       PointXY max,
                                       double ratio);
void       ifs_trans_scale            (IFSTrans *t,
                                       double r);
void       ifs_scale                  (IFS *ifs,
                                       double r);
void       ifs_trans_move             (IFSTrans *t,
                                       PointXY shift);
void       ifs_move                   (IFS *ifs,
                                       PointXY shift);
void       ifs_trans_rotate           (IFSTrans *t,
                                       double phi);
void       ifs_rotate                 (IFS *ifs,
                                       double phi);
void       ifs_trans_flip_horizontally(IFSTrans *t);
void       ifs_flip_horizontally      (IFS *ifs);
void       ifs_trans_flip_vertically  (IFSTrans *t);
void       ifs_flip_vertically        (IFS *ifs);
void       ifs_trans_asymmetrize      (IFSTrans *t,
                                       double alpha);
void       ifs_asymmetrize            (IFS *ifs,
                                       double alpha);
void       ifs_trans_skew             (IFSTrans *t,
                                       double k);
void       ifs_skew                   (IFS *ifs,
                                       double k);
void       ifs_trans_reflect          (IFSTrans *t,
                                       double omega);
void       ifs_reflect                (IFS *ifs,
                                       double omega);
void       ifs_trans_transform        (IFSTrans *t,
                                       const IFSTrans *tt);
void       ifs_transform              (IFS *ifs,
                                       const IFSTrans *tt);
IFSTrans*  ifs_trans_init_identity    (IFSTrans *t);
IFSTrans*  ifs_trans_init_rotation    (IFSTrans *t,
                                       double phi);
IFSTrans*  ifs_trans_init_scaling     (IFSTrans *t,
                                       double sx,
                                       double sy);
IFSTrans*  ifs_trans_init_skew        (IFSTrans *t,
                                       double k);
IFSTrans*  ifs_trans_compose          (const IFSTrans *t1,
                                       const IFSTrans *t2,
                                       IFSTrans *t);
IFSCTrans* ifs_trans_to_ifscompose    (const IFSTrans *t,
                                       PointXY center,
                                       IFSCTrans *ifsct);
IFSTrans*  ifs_trans_from_ifscompose  (const IFSCTrans *ifsct,
                                       PointXY center,
                                       IFSTrans *t);
IFS*       ifs_from_ifscompose        (const IFSCTrans *ifsc,
                                       size_t ntrans,
                                       PointXY center);
void       ifs_taus_seed              (void);
size_t*    ifs_discretize_probability (const IFS *ifs,
                                       size_t ntab,
                                       size_t *tab);
void       ifs_draw_int               (const IFS *ifs,
                                       size_t ntab,
                                       const size_t *tab,
                                       uint64_t niter,
                                       size_t rows,
                                       size_t cols,
                                       uint32_t *canvas);
uint32_t   ifs_int_get_max            (size_t rows,
                                       size_t cols,
                                       const uint32_t *canvas);
double     ifs_int_get_sqavg          (size_t rows,
                                       size_t cols,
                                       const uint32_t *canvas);
uint32_t*  ifs_int_get_histogram      (size_t rows,
                                       size_t cols,
                                       const uint32_t *canvas,
                                       uint32_t m);
uint32_t*  ifs_boxcount               (size_t rows,
                                       size_t cols,
                                       uint32_t *canvas,
                                       size_t maxsize);
double     ifs_estimate_dimension     (const IFS *ifs,
                                       size_t ntab,
                                       const size_t *tab);
void       ifs_estimate_dimension2    (const IFS *ifs,
                                       size_t ntab,
                                       const size_t *tab,
                                       double *dim,
                                       double *measure);
double     ifs_fit_boxcount_dimension (size_t max,
                                       const size_t *counts);
void       ifs_fit_boxcount_dimension2(size_t max,
                                       const size_t *counts,
                                       double *dim,
                                       double *measure);
size_t     ifs_int_get_nonzero        (size_t rows,
                                       size_t cols,
                                       const uint32_t *canvas);
void       ifs_int2gray8_converg      (unsigned char *pixmap,
                                       const uint32_t *canvas,
                                       size_t rows,
                                       size_t cols,
                                       uint32_t max,
                                       double qgamma);
void       ifs_int2gray16_converg     (uint16_t *pixmap,
                                       const uint32_t *canvas,
                                       size_t rows,
                                       size_t cols,
                                       uint32_t max,
                                       double qgamma);
void       ifs_int2gray8_bw           (unsigned char *pixmap,
                                       const uint32_t *canvas,
                                       size_t rows,
                                       size_t cols,
                                       uint32_t threshold);

#define IFS_TAUS_MASK 0xffffffffUL

/* this is not thread-safe if one minds what random number one exactly gets */
extern unsigned long int ifs_taus_state[4];

extern inline unsigned long
ifs_taus_get(void)
{
    unsigned long int b;

    b = (((ifs_taus_state[0] << 6UL) & IFS_TAUS_MASK) ^ ifs_taus_state[0])
        >> 13UL;
    ifs_taus_state[0]
        = (((ifs_taus_state[0] & 4294967294UL) << 18UL) & IFS_TAUS_MASK) ^ b;

    b = (((ifs_taus_state[1] << 2UL) & IFS_TAUS_MASK) ^ ifs_taus_state[1])
        >> 27UL;
    ifs_taus_state[1]
        = (((ifs_taus_state[1] & 4294967288UL) << 2UL) & IFS_TAUS_MASK) ^ b;

    b = (((ifs_taus_state[2] << 13UL) & IFS_TAUS_MASK) ^ ifs_taus_state[2])
        >> 21UL;
    ifs_taus_state[2]
        = (((ifs_taus_state[2] & 4294967280UL) << 7UL) & IFS_TAUS_MASK) ^ b;

    b = (((ifs_taus_state[3] << 3UL) & IFS_TAUS_MASK) ^ ifs_taus_state[3])
        >> 12UL;
    ifs_taus_state[3]
        = (((ifs_taus_state[3] & 4294967168UL) << 13UL) & IFS_TAUS_MASK) ^ b;

    return ifs_taus_state[0] ^ ifs_taus_state[1]
           ^ ifs_taus_state[2] ^ ifs_taus_state[3];
}

/* FIXME: this is here for gtk-doc, we do not provide any linkable version */
unsigned long ifs_taus_get(void);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
