/*
 *  @(#) $Id: taus.c,v 1.7 2004/04/06 12:11:11 yeti Exp $
 *  Copyright (C) 2004 David Necas (Yeti)
 *  E-mail: yeti@physics.muni.cz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */

#include "ifsgr.h"
#include <string.h>
#include <errno.h>

#ifndef RANDOM_FILE
#  define RANDOM_FILE "/dev/urandom"
#endif

#define IFS_TAUS_LCG(n) ((69069UL * n) & 0xffffffffUL)

/*
 * Based on Tausworthe random generator implementation rng/taus113.c
 * from the GNU Scientific Library (http://sources.redhat.com/gsl/)
 * which has the notice
 * Copyright (C) 2002 Atakan Gurkan
 * Based on the file taus.c which has the notice
 * Copyright (C) 1996, 1997, 1998, 1999, 2000 James Theiler, Brian Gough
 */

unsigned long int ifs_taus_state[4];

/**
 * ifs_taus_seed:
 *
 * Seed the Tausworthe random generator from /dev/urandom (or what was defined
 * as randomness source when IFSgr was compiled).  On failure 1 is used as
 * the default seed.
 **/
void
ifs_taus_seed(void)
{
    unsigned long int s;
    static const unsigned long int shift[] = { 2UL, 8UL, 16UL, 128UL };
    unsigned int i;
    FILE *fh;

    if (!(fh = fopen(RANDOM_FILE, "rb"))) {
        fprintf(stderr, "ifsgr_taus_seed: Cannot open %s: %s",
                RANDOM_FILE, strerror(errno));
        s = 1UL;
    }
    else {
        fread(&s, sizeof(unsigned long int), 1, fh);
        fclose(fh);
    }

    for (i = 0; i < 4; i++) {
        ifs_taus_state[i] = IFS_TAUS_LCG(i ? ifs_taus_state[i-1] : s);
        if (ifs_taus_state[i] < shift[i])
            ifs_taus_state[i] += shift[i];
    }

    /* Calling RNG ten times to satify recurrence condition */
    for (i = 0; i < 10; i++)
        ifs_taus_get();
}

/*********** Documentation ***************************************************/
/**
 * ifs_taus_get:
 *
 * Get next random number.
 *
 * This function is not thread safe.  As there is no support for explicite
 * seeding (yet?), one random number is like another.
 *
 * Returns: A random number in the range 0..@IFS_TAUS_MASK.
 **/

/**
 * IFS_TAUS_MASK:
 *
 * The maximum value ifs_taus_get() can return, and a bit mask for numbers
 * ifs_taus_get() can return at the same time.
 **/

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
